﻿using System;
using System.Collections.Generic;
using System.Threading;
using System.Windows;
using ThinkGeo.MapSuite;
using ThinkGeo.MapSuite.Drawing;
using ThinkGeo.MapSuite.Layers;
using ThinkGeo.MapSuite.Shapes;
using ThinkGeo.MapSuite.Styles;
using ThinkGeo.MapSuite.Wpf;

namespace TestCollectionWasModified
{
    public static class NumericExtensions
    {
        public static double ToRadians(this double val)
        {
            return (Math.PI / 180.0) * val;
        }
    }
    public class TxEntry
    {
        public int Id;
        public string Name;
        public double Latitude;
        public double Longitude;
        public double Radius;
        public double Bearing;
        public bool Selected;
    }
    public class EntryChangedEventArgs<T> : EventArgs
    {
        public T entry;
        public T old_entry;
        public EntryChangedEventArgs(T entry)
        {
            this.entry = entry;
            old_entry = entry;
        }
        public EntryChangedEventArgs(T entry, T old_entry)
        {
            this.entry = entry;
            this.old_entry = old_entry;
        }
    }
    public partial class MainWindow : Window
    {
        enum MapChoices
        {
            WorldStreetsLayer,
            WorldStreetsAndImageryOverlay,
            OpenStreetMapOverlay,
            End
        }

        MapChoices map_choice = MapChoices.OpenStreetMapOverlay;
        int num_entries = 100;
        int refresh_delay = 100; // TODO set this value to 10 in order to replicate the crash in Map.Refresh
        Proj4Projection proj4 = null;
        List<TxEntry> tx_entries;
        InMemoryFeatureLayer tx_selected_layer;
        InMemoryFeatureLayer tx_unselected_layer;
        LayerOverlay tx_overlay;
        EditInteractiveOverlay tx_label_overlay;

        public MainWindow()
        {
            InitializeComponent();
        }
        private void MapLoaded(object sender, RoutedEventArgs e)
        {
            Map.Loaded -= MapLoaded;
            string countries_file = "..\\..\\AppData\\Countries02.shp";
            if (!System.IO.File.Exists(countries_file))
                throw new Exception("Countries file " + countries_file + " does not exist");

            string data_file = "..\\..\\AppData\\DallasCounty-3857-20170218.sqlite";
            if (map_choice == MapChoices.WorldStreetsLayer && !System.IO.File.Exists(data_file))
                throw new Exception("Data file " + data_file + " does not exist");

            string cache_directory = "..\\..\\AppData";
            string cache_id = "Cache";

            AreaStyle area_style = new AreaStyle();
            area_style.FillSolidBrush = new GeoSolidBrush(GeoColor.FromArgb(255, 233, 232, 214));
            area_style.OutlinePen = new GeoPen(GeoColor.FromArgb(255, 118, 138, 69), 1);
            area_style.OutlinePen.DashStyle = LineDashStyle.Solid;

            ShapeFileFeatureLayer world_layer = new ShapeFileFeatureLayer(countries_file);
            world_layer.ZoomLevelSet.ZoomLevel01.DefaultAreaStyle = area_style;
            world_layer.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;

            LayerOverlay overlay = new LayerOverlay();
            overlay.Layers.Add(new BackgroundLayer(new GeoSolidBrush(GeoColor.GeographicColors.ShallowOcean)));
            overlay.Layers.Add(world_layer);
            Map.Overlays.Add(overlay);

            switch (map_choice)
            {
                /* case MapChoices.WorldStreetsLayer:
                    Map.MapUnit = GeographyUnit.Meter;
                    WorldStreetsLayer world_streets_layer = new WorldStreetsLayer("Data Source=" + data_file + ";Version=3;Read Only=True;");
                    LayerOverlay layer_overlay = new LayerOverlay();
                    layer_overlay.Layers.Add(world_streets_layer);
                    Map.Overlays.Add("World Streets Layer", layer_overlay);
                    break; */
                case MapChoices.WorldStreetsAndImageryOverlay:
                    Map.MapUnit = GeographyUnit.DecimalDegree;
                    WorldStreetsAndImageryOverlay world_streets_and_imagery_overlay = new WorldStreetsAndImageryOverlay();
                    world_streets_and_imagery_overlay.Name = "World Streets And Imagery Overlay";
                    world_streets_and_imagery_overlay.TileCache = new FileBitmapTileCache(cache_directory, cache_id);
                    Map.Overlays.Add(world_streets_and_imagery_overlay);
                    break;
                case MapChoices.OpenStreetMapOverlay:
                    Map.MapUnit = GeographyUnit.Meter;
                    OpenStreetMapOverlay open_street_map_overlay = new OpenStreetMapOverlay();
                    open_street_map_overlay.Name = "Open Street Map Overlay";
                    open_street_map_overlay.TileCache = new FileBitmapTileCache(cache_directory, cache_id);
                    Map.Overlays.Add(open_street_map_overlay);
                    break;
                default:
                    throw new Exception("Unrecognised map choice");
            }

            proj4 = new Proj4Projection(Proj4Projection.GetDecimalDegreesParametersString(), Proj4Projection.GetBingMapParametersString()); // TODO what projection settings should be used for the different kinds of map data?
            proj4.Open();

            tx_entries = new List<TxEntry>(num_entries);
            for (int i = 0; i < num_entries; i++)
            {
                TxEntry tx_entry = new TxEntry();
                tx_entry.Id = i;
                tx_entry.Name = "Wibble" + i;
                tx_entry.Selected = true;
                tx_entries.Add(tx_entry);
            }

            TxMap();
            TxDisplay();

            new Thread(new ThreadStart(() =>
            {
                ModifyTxEntries();
            })).Start();
        }
        private void TxMap()
        {
            tx_selected_layer = new InMemoryFeatureLayer();
            tx_selected_layer.ZoomLevelSet.ZoomLevel01.DefaultLineStyle = LineStyles.CreateSimpleLineStyle(GeoColors.Red, 1, true);
            tx_selected_layer.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;

            tx_unselected_layer = new InMemoryFeatureLayer();
            tx_unselected_layer.ZoomLevelSet.ZoomLevel01.DefaultLineStyle = LineStyles.CreateSimpleLineStyle(GeoColors.Black, 1, true);
            tx_unselected_layer.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;

            tx_overlay = new LayerOverlay();
            tx_overlay.Layers.Add("TxSelectedLayer", tx_selected_layer);
            tx_overlay.Layers.Add("TxUnselectedLayer", tx_unselected_layer);
            Map.Overlays.Add("TxOverlay", tx_overlay);

            tx_label_overlay = new EditInteractiveOverlay();
            tx_label_overlay.EditShapesLayer.ZoomLevelSet.ZoomLevel01.DefaultAreaStyle.IsActive = false;
            tx_label_overlay.EditShapesLayer.ZoomLevelSet.ZoomLevel01.DefaultLineStyle.IsActive = false;
            tx_label_overlay.EditShapesLayer.ZoomLevelSet.ZoomLevel01.DefaultPointStyle.IsActive = false;
            tx_label_overlay.EditShapesLayer.ZoomLevelSet.ZoomLevel01.DefaultTextStyle.IsActive = false;
            tx_label_overlay.EditShapesLayer.Open();
            tx_label_overlay.EditShapesLayer.Columns.Add(new FeatureSourceColumn("Label"));
            tx_label_overlay.EditShapesLayer.Close();
            Map.InteractiveOverlays.Add("TxLabels", tx_label_overlay);
        }
        private void TxDisplay()
        {
            tx_selected_layer.InternalFeatures.Clear();
            tx_unselected_layer.InternalFeatures.Clear();
            tx_label_overlay.EditShapesLayer.InternalFeatures.Clear();

            foreach (TxEntry tx_entry in tx_entries)
                TxAdd(tx_entry);

            TxRefresh();
        }
        private void ButtonClick(object sender, RoutedEventArgs e)
        {
            foreach (TxEntry tx_entry in tx_entries)
            {
                EntryChangedEventArgs<TxEntry> event_args = new EntryChangedEventArgs<TxEntry>(tx_entry);
                TxEntryChangedEventHandler(sender, event_args);
            }
        }
        private void TxEntryChangedEventHandler(object sender, EntryChangedEventArgs<TxEntry> event_args)
        {
            Dispatcher.Invoke((Action)(() => { TxChange(event_args.entry); }));
            TxRefresh();
        }
        private void TxChange(TxEntry tx_entry)
        {
            TxDelete(tx_entry);
            TxAdd(tx_entry);
        }
        private void TxDelete(TxEntry tx_entry)
        {
            if (tx_selected_layer.InternalFeatures.Count > 0)
                tx_selected_layer.InternalFeatures.Remove(tx_entry.Id.ToString());
            if (tx_unselected_layer.InternalFeatures.Count > 0)
                tx_unselected_layer.InternalFeatures.Remove(tx_entry.Id.ToString());
            if (tx_label_overlay.EditShapesLayer.InternalFeatures.Count > 0)
                tx_label_overlay.EditShapesLayer.InternalFeatures.Remove(tx_entry.Id.ToString());
        }
        private void TxAdd(TxEntry tx_entry)
        {
            double horiz_offset = tx_entry.Radius * Math.Sin(tx_entry.Bearing.ToRadians());
            double vert_offset = tx_entry.Radius * Math.Cos(tx_entry.Bearing.ToRadians());
            List<Vertex> vertices = new List<Vertex>();
            vertices.Add(new Vertex(tx_entry.Longitude, tx_entry.Latitude));
            vertices.Add(new Vertex(tx_entry.Longitude + horiz_offset, tx_entry.Latitude + vert_offset));
            LineShape line = new LineShape(vertices);
            line.Tag = tx_entry.Id;

            TextStyle text_style = new TextStyle("Label", new GeoFont("Arial", 10), new GeoSolidBrush(GeoColor.SimpleColors.Black));
            text_style.PointPlacement = PointPlacement.Center;
            ValueStyle value_style = new ValueStyle();
            value_style.ColumnName = "Label";
            value_style.ValueItems.Add(new ValueItem(tx_entry.Name, text_style));
            tx_label_overlay.EditShapesLayer.ZoomLevelSet.ZoomLevel01.CustomStyles.Add(value_style);
            tx_label_overlay.EditShapesLayer.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;

            ColourEntry(tx_entry, tx_selected_layer, tx_unselected_layer, tx_label_overlay, line, add_label: false);

            PointShape point = new PointShape(tx_entry.Longitude + horiz_offset, tx_entry.Latitude + vert_offset);
            ColourEntry(tx_entry, tx_selected_layer, tx_unselected_layer, tx_label_overlay, point, add_label: true);
        }
        private void ColourEntry<T>(TxEntry entry, InMemoryFeatureLayer layer_selected, InMemoryFeatureLayer layer_unselected, EditInteractiveOverlay label_overlay, T shape, bool add_label) where T : BaseShape
        {
            if (Map.MapUnit == GeographyUnit.Meter)
                shape = (T)proj4.ConvertToExternalProjection(shape); // TODO is this the correct conversion?

            Feature feature = new Feature(shape);
            feature.ColumnValues["Label"] = entry.Name;

            if (entry.Selected)
            {
                if (!layer_selected.InternalFeatures.Contains(entry.Id.ToString()))
                    layer_selected.InternalFeatures.Add(entry.Id.ToString(), feature);
                if (layer_unselected.InternalFeatures.Count > 0)
                    layer_unselected.InternalFeatures.Remove(entry.Id.ToString());
                if (add_label && !label_overlay.EditShapesLayer.InternalFeatures.Contains(entry.Id.ToString()))
                    label_overlay.EditShapesLayer.InternalFeatures.Add(entry.Id.ToString(), feature);
            }
            else
            {
                if (!layer_unselected.InternalFeatures.Contains(entry.Id.ToString()))
                    layer_unselected.InternalFeatures.Add(entry.Id.ToString(), feature);
                if (layer_selected.InternalFeatures.Count > 0)
                    layer_selected.InternalFeatures.Remove(entry.Id.ToString());
                if (add_label && label_overlay.EditShapesLayer.InternalFeatures.Contains(entry.Id.ToString()))
                    label_overlay.EditShapesLayer.InternalFeatures.Remove(entry.Id.ToString());
            }
        }
        private void TxRefresh()
        {
            List<Overlay> overlays = new List<Overlay>();
            overlays.Add(tx_overlay);
            overlays.Add(tx_label_overlay);
            Dispatcher.Invoke((Action)(() => { Map.Refresh(overlays); }));
            Thread.Sleep(refresh_delay); // TODO adding this sleep with a sufficiently long delay is the workaround for the crash in Map.Refresh
        }
        private void ModifyTxEntries()
        {
            Random random = new Random();
            while (true)
            {
                foreach (TxEntry tx_entry in tx_entries)
                {
                    tx_entry.Latitude = random.Next(-90, 90);
                    tx_entry.Longitude = random.Next(-180, 180);
                    tx_entry.Radius = random.Next(1, 10);
                    tx_entry.Bearing = random.Next(0, 360);
                    Dispatcher.Invoke((Action)(() => { TxChange(tx_entry); }));
                    TxRefresh();
                }
            }
        }
    }
}
